import numpy as np
from scipy.stats import binom

### Private functions
def _jc_posterior_ng(COUNT_MATRIX, DIST_SAMPLES):
    MATRIX_SUM = COUNT_MATRIX.sum()
    MATRIX_DIAGONAL_SUM = COUNT_MATRIX.diagonal().sum()
    P = np.exp(- DIST_SAMPLES / 100)
    
    likelihood = binom.pmf(MATRIX_DIAGONAL_SUM, MATRIX_SUM, P)

#   This code is not commented in Octave
#    if (any(isnan(likelihood)))
#      # In case the binomial is tricky to compute, approx with normal distr!
#      likelihood = normpdf(k, tot .* p, tot .* p .* (1 .- p)); 
#    endif
    
    likelihood[0] /= 2
    likelihood[-1] /= 2
    
    POSTERIOR_VEC = likelihood / ( likelihood.sum() * (DIST_SAMPLES[1] - DIST_SAMPLES[0]) )

    return POSTERIOR_VEC    
    
### Interface
def comp_posterior_JC(COUNT_MATRIX_LIST, DIST_SAMPLES):
    NUMBER_OF_COUNT_MATRICES = len(COUNT_MATRIX_LIST)
    NUMBER_OF_DIST_SAMPLES = len(DIST_SAMPLES)
    PD = np.empty((NUMBER_OF_COUNT_MATRICES, NUMBER_OF_DIST_SAMPLES))
    
    PD = np.array([_jc_posterior_ng(COUNT_MATRIX, DIST_SAMPLES) for COUNT_MATRIX in COUNT_MATRIX_LIST])
    return PD
        


import numpy as np
from scipy.linalg import eig

def find_zero_eigenvalue_eigenvector(matrix):
    '''
    Find the index of the eigenvector corresponding to Q's zero eigenvalue.
    This is recognized as the row (because we will be looking at the 'right'
    eigenvectors, not the usual left) with all positive or all negative elements.
    '''
    zero_eigenvector_candidates = []
    for index, eigen_vector in enumerate(matrix):
        if all(eigen_vector > 0) or all(eigen_vector < 0):
         zero_eigenvector_candidates.append((eigen_vector, index))

    if (len(zero_eigenvector_candidates) > 1):
        raise ValueError("More than one candidate for null-vector")
    if (len(zero_eigenvector_candidates) == 0):
        raise ValueError("No candidate for null-vector!")
        
    eigen_vector, index = zero_eigenvector_candidates[0]

    #   Returns as list to return copy and not reference
    return list(eigen_vector), index


def find_eigens(count_matrix_list):
    p_sum = sum(0.5 * (matrix + matrix.T) for matrix in count_matrix_list)

    # Make every row sum to 1
    row_sums = np.linalg.norm(p_sum, axis=1, ord=1, keepdims=1)
    p_sum = np.divide(p_sum, row_sums, out=np.zeros_like(p_sum), where=row_sums != 0)   # Only divide where the row sum is non-zero

    try:
        eigen_values, vr = eig(p_sum, left=False, right=True)   #   Calculate eigenvalues and the right eigenvectors of p_sum
    except ValueError:
        raise ValueError("Unable to calculate eigenvalues")

    if not np.all(np.isreal(eigen_values)):
        raise ValueError("An eigenvalue is complex")

    vl = np.linalg.inv(vr)

    eq,_ = find_zero_eigenvalue_eigenvector(vl)
    eq /= np.linalg.norm(eq, ord=1)
    eq = np.absolute(eq)

    return vl, vr, eq


# Scale Q so that the average mutation rate is 0.01
def scale_q(Q, EQ):
        SCALE_FACTOR = np.dot(EQ, (-np.diag(Q)))

        if(SCALE_FACTOR == 0):
            raise ZeroDivisionError('No Q diagonal cause a problem in estimate_q.py:scale_q')

        Q /= SCALE_FACTOR

        return Q


#   Sometimes, when data is sparse, Q estimates come out with
#   off-diagonal entries being negative. Not good.
def fix_negatives(Q):
    # Replace negative elements with smallest absolute value
    MINIMUM_ELEMENT = np.min(np.abs(Q))
    Q[Q<0] = MINIMUM_ELEMENT
    #   Recalculate diagonal to be negative rowsums
    np.fill_diagonal(Q, 0)
    ROW_SUMS = Q.sum(axis=1)
    np.fill_diagonal(Q, -ROW_SUMS)

    return Q


def recover_q(L, VR, VL):
    Q = 0.01 * (VR @ np.diag(L) @ VL)
    return Q

# Alternative: Estimate eigenvalues of Q using weighted points
#
# The estimated eigenvalues are returned in L
def _weighted_estimate_eigenvals(PW, W, VL, VR, DIST_SAMPLES):
    # The X and Y matrises will contain 20 least-squares problems, one for each eigenvalue
    NUMBER_OF_DIST_SAMPLES = len(DIST_SAMPLES)
    X = np.empty(NUMBER_OF_DIST_SAMPLES, dtype="float64").reshape((80,1))
    Y = np.empty((20, NUMBER_OF_DIST_SAMPLES))

    # Find the eigenvector corresponding to eigenvalue = 1
    _, NULL_VECTOR_INDEX = find_zero_eigenvalue_eigenvector(VL)

    # Gather some datapoints
    for i, DIST_SAMPLE in enumerate(DIST_SAMPLES):
        ELAMBDA = np.diag(VL @ PW[i] @ VR)
        X[i] = (DIST_SAMPLE / 100) * W[i]

        for li in range(20):
            if (li == NULL_VECTOR_INDEX):
                continue

            if (ELAMBDA[li] > 0):    # Skip complex value data points!
                Y[li, i] = np.real(np.log(ELAMBDA[li])) * W[i]
            else:
                X[i] = 0   # No disturbance if set to 0!
                Y[li, i] = 0

    L = np.zeros(20)

    for i,_ in enumerate(L):
        if(i == NULL_VECTOR_INDEX):
            L[i] = 0
        else:
            tempY = Y[i,:].reshape(80,1)
            res = np.linalg.lstsq(X, tempY, rcond = None)[0][0][0]
            L[i] = res

    return L


### Interface
def estimate_q(PW, W, VL, VR, EQ, DIST_SAMPLES):
    L = _weighted_estimate_eigenvals(PW, W, VL, VR, DIST_SAMPLES)

    Q = recover_q(L, VR, VL)
    Q = fix_negatives(Q)
    Q = scale_q(Q, EQ)

    return Q



import copy
import random
import numpy as np
import math
import sys
import os

### TO BE DELETED

sys.path.insert(1, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from _bw_estimator.bw_estimator import bw_estimator

#   Private functions
def resample_columns(MULTIALIGNMENT_ARRAY):
    new_multialignment = np.empty_like(MULTIALIGNMENT_ARRAY)
    SEQUENCE_LENGTH = MULTIALIGNMENT_ARRAY.shape[1]

    for COLUMN_INDEX in range(SEQUENCE_LENGTH):
        RANDOM_INDEX = random.randint(0, SEQUENCE_LENGTH - 1)
        new_multialignment[:, COLUMN_INDEX] = MULTIALIGNMENT_ARRAY[:, RANDOM_INDEX]

    return new_multialignment

def calculate_bw_for_resamplings(resamplings, threshold, multialignment):
    q_list = []
    eq_list = []
    number_of_times_bw_estimator_failed = 0

    for _ in range(resamplings):
        resampled_multialignment = resample_columns(multialignment)        
        resampled_multialignment_list = [resampled_multialignment]

        try:
            Q, EQ = bw_estimator(threshold, resampled_multialignment_list)
            q_list.append(Q)
            eq_list.append(EQ)
        except:
            number_of_times_bw_estimator_failed +=1

    failed_percentage = number_of_times_bw_estimator_failed / resamplings
    return q_list, failed_percentage

def q_diff_mean(REFERENCE_Q, RESAMPLED_Q_LIST):
    Q_DIFF_NORM_LIST = []

    for Q in RESAMPLED_Q_LIST:
        Q_DIFF = REFERENCE_Q - Q
        Q_DIFF_NORM = np.linalg.norm(Q_DIFF)
        Q_DIFF_NORM_LIST.append(Q_DIFF_NORM)
    Q_DIFF_MEAN = np.mean(Q_DIFF_NORM_LIST)

    return Q_DIFF_MEAN

def q_bootstrap_estimate(resampled_Q_list):
    '''
    Compute the mean Q matrix and elementwise standard deviations given a list of bootstrap Q
    estimate from resampled alignments.

    Return estimated Q and the std dev as an error estimate.
    '''
    collection = np.stack(resampled_Q_list)
    mean_estimate = np.mean(collection, axis=0) # Corresponds to elementwise mean
    std_dev = np.std(collection, axis=0)        # ...and standard deviation

    return mean_estimate, std_dev


#   Interface
def bootstrapper(resamplings, threshold, multialignment):
    multialignment_list = [multialignment]
    try:
        reference_Q, _ = bw_estimator(threshold, multialignment_list)
    except:
        raise ValueError("Failed to estimate a baseline Q matrix in bootstrap procedure")

    resampled_q_list, failed_percentage = calculate_bw_for_resamplings(resamplings, threshold, multialignment)
    mean_difference = q_diff_mean(reference_Q, resampled_q_list)
    mean_difference *= 10000

    return mean_difference, failed_percentage



import tempfile
import numpy as np
import os
import sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))
from modelestimator._bw_estimator.bw_estimator import bw_estimator
from modelestimator.io import handle_input_file

def test_case_2(tmpdir):
        #   Create directory paths
    CURRENT_DIR = os.path.dirname(__file__)
    TEST_FILES_PATH = os.path.join(CURRENT_DIR, 'test_case_2/')

    #   Create sequence file path
    SEQUENCE_FILE_NAME = "testcase2_20seqs_1000long_50pam.fa"
    FILE_PATH = os.path.join(TEST_FILES_PATH, SEQUENCE_FILE_NAME)

    #   Load reference Q and EQ
    REFERENCE_Q_PATH = os.path.join(TEST_FILES_PATH, 'reference_Q.txt')
    REFERENCE_Q = np.loadtxt(REFERENCE_Q_PATH)
    REFERENCE_EQ_PATH = os.path.join(TEST_FILES_PATH, 'reference_EQ.txt')
    REFERENCE_EQ = np.loadtxt(REFERENCE_EQ_PATH)

    #   Calculate Q and EQ
    FORMAT = "fasta"
    MULTIALIGNMENT = handle_input_file(FILE_PATH, FORMAT)
    MULTLIGNMENT_LIST = [MULTIALIGNMENT]
    THRESHOLD = 0.001
    CALCULATED_Q, CALCULATED_EQ = bw_estimator(THRESHOLD, MULTLIGNMENT_LIST)

    #   Assert that calculated and references are close. Expected to pass
    assert(np.allclose(CALCULATED_Q, REFERENCE_Q, atol=THRESHOLD))
    assert(np.allclose(CALCULATED_EQ, REFERENCE_EQ, atol=THRESHOLD))



import os
import sys
import numpy as np
from Bio import AlignIO

sys.path.insert(1, os.path.abspath(os.path.join(sys.path[0], "..")))
from modelestimator.bootstrap import q_bootstrap_estimate
from modelestimator.io import handle_input_file

def test_bootstrap_1():
    REFERENCE_FILE_PATH = os.path.abspath(os.path.join(sys.path[0], "tests/test_bootstrap_1/1000LongMultialignment.phylip"))
    MULTIALIGNMENT = handle_input_file(REFERENCE_FILE_PATH, "phylip")

    RESAMPLINGS = 25
    THRESHOLD = 0.001
    Q_mean, EQ_mean, Q_SD, n_failures = q_bootstrap_estimate([MULTIALIGNMENT], THRESHOLD, RESAMPLINGS)
    assert n_failures == 0
    assert np.sum(Q_SD) < 190*0.2

test_bootstrap_1()



from ._calculate_q_eq.match_closest_pair import choose_close_pairs
from ._calculate_q_eq.create_count_matrices import create_count_matrices
from ._calculate_q_eq.calculate_q_eq import calculate_q_eq

def bw_estimator(threshold, msa_list, compare_indels_flag = False):
    '''
    threshold  -  Decides when to stop iterations
    msa_list   -  A list of alignments 
    compare_indels_flag  -  decides if indels should be included when comparing likeness of sequences

    Returns: a rate matrix and a corresponding steady state distribution on amino acids,
             as estimated with the BW method.
    '''
    aggregated_count_matrix_list = []
    
    for msa in msa_list:
        close_pairs = choose_close_pairs(msa, compare_indels_flag)
        count_matrix_list = create_count_matrices(close_pairs)
        aggregated_count_matrix_list.extend(count_matrix_list)

    Q, eq = calculate_q_eq(aggregated_count_matrix_list, threshold)
    
    return Q, eq




import copy
import random
import numpy as np
import math
import sys
import os

from ._bw_estimator.bw_estimator import bw_estimator
from ._bw_estimator._calculate_q_eq.match_closest_pair import choose_close_pairs
from ._bw_estimator._calculate_q_eq.create_count_matrices import create_count_matrices
from ._bw_estimator._calculate_q_eq.calculate_q_eq import calculate_q_eq

#   Private functions
def resample_columns(multialignment_array):
    new_multialignment = np.empty_like(multialignment_array)
    sequence_length = multialignment_array.shape[1]

    for column_index in range(sequence_length):
        random_index = random.randint(0, sequence_length - 1)
        new_multialignment[:, column_index] = multialignment_array[:, random_index]

    return new_multialignment

def calculate_bw_for_resamplings(resamplings, threshold, multialignment):
    q_list = []
    eq_list = []
    number_of_times_bw_estimator_failed = 0

    for _ in range(resamplings):
        resampled_multialignment = resample_columns(multialignment)        
        resampled_multialignment_list = [resampled_multialignment]

        try:
            Q, EQ = bw_estimator(threshold, resampled_multialignment_list)
            q_list.append(Q)
            eq_list.append(EQ)
        except:
            number_of_times_bw_estimator_failed +=1

    failed_percentage = number_of_times_bw_estimator_failed / resamplings
    return q_list, failed_percentage

def q_diff_mean(REFERENCE_Q, RESAMPLED_Q_LIST):
    Q_DIFF_NORM_LIST = []

    for Q in RESAMPLED_Q_LIST:
        Q_DIFF = REFERENCE_Q - Q
        Q_DIFF_NORM = np.linalg.norm(Q_DIFF)
        Q_DIFF_NORM_LIST.append(Q_DIFF_NORM)
    Q_DIFF_MEAN = np.mean(Q_DIFF_NORM_LIST)

    return Q_DIFF_MEAN


#   Interface
def bootstrapped_stability(resamplings, threshold, multialignment):
    multialignment_list = [multialignment]
    try:
        reference_Q, _ = bw_estimator(threshold, multialignment_list)
    except:
        raise ValueError("Failed to estimate a baseline Q matrix in bootstrap procedure")

    resampled_q_list, failed_percentage = calculate_bw_for_resamplings(resamplings, threshold, multialignment)
    mean_difference = q_diff_mean(reference_Q, resampled_q_list)
    mean_difference *= 10000

    return mean_difference, failed_percentage


def q_bootstrap_estimate(msa_list, threshold, n_bootstraps, compare_indels_flag=False):
    '''
    Compute the mean Q matrix and elementwise standard deviations from resamplings of an input alignment.

    Return estimated Q and the std dev as an error estimate.
    '''
    resampled_Q_list = []
    mean_eq = None
    n_failures = 0
    for i in range(n_bootstraps):
        try:
            aggregated_count_matrix_list = []
            for msa in msa_list:
                resampled_msa = resample_columns(msa)
                close_pairs = choose_close_pairs(resampled_msa, compare_indels_flag)
                count_matrices = create_count_matrices(close_pairs)
                aggregated_count_matrix_list.extend(count_matrices)
            Q, eq = calculate_q_eq(aggregated_count_matrix_list, threshold)
            resampled_Q_list.append(Q)
            if i > 0:
                mean_eq += eq
            else:
                mean_eq = eq
        except Exception as e:
            print('Bootstrap replicate problem:', str(e))
            n_failures += 1

    failed_percentage = float(n_failures) / n_bootstraps
    if failed_percentage <= 0.25:
        collection = np.stack(resampled_Q_list)
        mean_estimate = np.mean(collection, axis=0) # Corresponds to elementwise mean
        mean_eq = mean_eq / (n_bootstraps - n_failures)
        std_dev = np.std(collection, axis=0)        # ...and standard deviation
        return mean_estimate, mean_eq, std_dev, n_failures
    else:
        raise ValueError(f'Too many failed bootstrap replicate attempts, {n_failures} out of {n_bootstraps} (i.e., {failed_percentage:.2})')




from .comp_posterior import comp_posterior
from .matrix_weight import matrix_weight
from .estimate_q import estimate_q

import numpy as np

def simple_estimation(COUNT_MATRIX_LIST, Q_OLD, VL, VR, EQ, DIST_SAMPLES):
    POSTERIOR  = comp_posterior(COUNT_MATRIX_LIST, Q_OLD, EQ, DIST_SAMPLES)
    W = POSTERIOR.sum(axis=0)
    PW = matrix_weight(COUNT_MATRIX_LIST, POSTERIOR, DIST_SAMPLES)
    Q_NEW = estimate_q(PW, W, VL, VR, EQ, DIST_SAMPLES)
    
    return Q_NEW


import os
import sys
import numpy as np
from .comp_posterior_JC import comp_posterior_JC
from .matrix_weight import matrix_weight
from .estimate_q import estimate_q, find_zero_eigenvalue_eigenvector, find_eigens
from .simple_estimation import simple_estimation


def calculate_q_eq(count_matrix_list, threshold):
    #   VL = Inverse of right eigenvectorts
    #   VR = Right eigenvectors
    #   EQ = Right eigenvector corresponding to zero eigenvalue but normalized so it sums to 1
    vl, vr, eq = find_eigens(count_matrix_list)

    #   Get a first simple estimate of Q using a Jukes-Cantor model
    dist_samples = np.arange(1, 400, 5)
    posterior = comp_posterior_JC(count_matrix_list, dist_samples)   # posterior.shape = (10, 80). Rows are identical to Octave but in different order
    pw = matrix_weight(count_matrix_list, posterior, dist_samples)    
    w = posterior.sum(axis=0)
    q = estimate_q(pw, w, vl, vr, eq, dist_samples)

    #   Set loop variables
    difference = 1+threshold
    iterations = 0
    max_iterations = 10
    
    #   Calculate Q
    while (iterations < max_iterations and difference > threshold):
        iterations += 1
        q_new = simple_estimation(count_matrix_list, q, vl, vr, eq, dist_samples)
        difference = np.linalg.norm(q_new - q)
        q = q_new
    return q, eq



import os
import sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))
from modelestimator._bw_estimator.bw_estimator import bw_estimator
from modelestimator.io import handle_input_file

#   Test that bw_estimator works with a multialignment
#   consisting of an odd number of sequences

def test_odd_sequences():
    #   Create directory paths
    CURRENT_DIR = os.path.dirname(__file__)
    FOLDER_DIR = 'test_odd_sequences/'
    TEST_FILES_PATH = os.path.join(CURRENT_DIR, FOLDER_DIR)

    #   Create sequence file path
    SEQUENCE_FILE_NAME = "nad2.fa.pep.aln"
    FILE_PATH = os.path.join(TEST_FILES_PATH, SEQUENCE_FILE_NAME)

    #   Test run
    FORMAT = "fasta"
    MULTIALIGNMENT = handle_input_file(FILE_PATH, FORMAT)
    MULTIALIGNMENT_LIST = [MULTIALIGNMENT]
    THRESHOLD = 0.001
    _,_ = bw_estimator(THRESHOLD, MULTIALIGNMENT_LIST)



import sys
import numpy as np
from Bio import AlignIO

def handle_input_file(file_path, format):
    '''
    Return aligned sequences found in given file.
    Note that the format parameter is a string accepted by AlignIO.read().
    '''
    seq_list = AlignIO.read(file_path, format)
    seq_list = [str(sequence.seq) for sequence in seq_list]

    number_of_sequences = len(seq_list)
    sequence_length = len(seq_list[0])
    new_array = np.empty([number_of_sequences, sequence_length], dtype='U')

    for index, sequence in enumerate(seq_list):
        new_array[index] = np.array(list(sequence))

    return new_array

def format_model_output(Q, eq, out_format):
    if out_format in ['iqtree', 'paml', 'phyml', 'raxml']:
        return paml_model_output(Q, eq)
    elif out_format == 'mrbayes':
        return mrbayes_model_output(Q, eq)
    elif out_format in ['matlab', 'octave']:
        return octave_model_output(Q, eq)

def paml_model_output(Q, eq):
    '''
    Construct and return a string containing the
    R matrix (which is symmetric and q_ij = r_ij*p_j)
    for subsequent use in PAML.
    '''
    output_string = ''
    for row in range(1,20):
        for col in range(0, row):
            r = Q[row, col] / eq[col]
            output_string += f'{r:<8.5} '
        output_string = output_string.rstrip()
        output_string += '\n'

    eq_str = ''
    for elem in range(0,20):
        eq_str += f'{eq[elem]:<8.5} '

    output_string += '\n' + eq_str.rstrip()
    return output_string

def mrbayes_model_output(Q, eq):
    '''
    Output the MrBayes instruction to get user defined rates.

    WARNING: This output is not verified against MrBayes.
    '''
    output_string = 'prset aarevmatpr = fixed('
    rate_list = []
    for row in range(1,20):
        for col in range(0, row):
            r = Q[row, col] / eq[col]
            rate_list.append(f'{r:8.5}')
    output_string += ','.join(rate_list) + ')'
    return output_string

def octave_model_output(Q, eq):
    '''
    Output an Octave-style matrix and vector assigned to 'Q' and a 'eq'.
    NOTE: It is the Q matrix, not the symmetric replacment rates as e.g. PAML
          and PhyML prefers.
    '''
    q_str = '''# name: Q
# type: matrix
# rows: 20
# columns: 20
'''
    for row in Q:
        q_str += '  '.join(map(lambda e: str(round(e, 6)),row))
        q_str += '\n'

    eq_str = '''# name: eq
# type: matrix
# rows: 1
# columns: 20
'''
    eq_str += ' '.join(map(lambda e: str(round(e, 6)), eq))
    return q_str + '\n' + eq_str

# ...but if outputting the R matrix (which is symmetric and q_ij = r_ij*p_j)
# we get compatibility with other tools, like PhyML.



import numpy as np

### Private functions
def _create_count_matrix(SEQUENCE_PAIRS):
    return_matrix = np.zeros((20,20))
    
    ALPHABET = 'ARNDCQEGHILKMFPSTWYV'
    alphabet_dictionary = {}
    for i, letter in enumerate(ALPHABET):
        alphabet_dictionary[letter] = i
    
    for i,_ in enumerate(SEQUENCE_PAIRS[0]):
        a = SEQUENCE_PAIRS[0][i]
        b = SEQUENCE_PAIRS[1][i]
        
        if a not in ALPHABET or b not in ALPHABET:
            continue
        
        return_matrix[alphabet_dictionary[b]][alphabet_dictionary[a]] = return_matrix[alphabet_dictionary[b]][alphabet_dictionary[a]] + 1
        
    return return_matrix

### Interface
def create_count_matrices(SEQUENCE_PAIRS):
    NUMBER_OF_SEQUENCE_PAIRS = len(SEQUENCE_PAIRS)
    count_matrix_list = np.empty((NUMBER_OF_SEQUENCE_PAIRS, 20, 20))

    count_matrix_list = np.array([_create_count_matrix(SEQUENCE_PAIR) for SEQUENCE_PAIR in SEQUENCE_PAIRS])
    
    return count_matrix_list



import numpy as np
import scipy

### Private functions

# Addition of log-probabilities
def _logprob_add(P, Q):
    if (P < Q):
        log_ratio = Q + np.log(1 + np.exp(P-Q))
    else:
        log_ratio = P + np.log(1 + np.exp(Q-P))
        
    return log_ratio

def _log_lik(COUNT_MATRIX, PT):
    P = np.sum(COUNT_MATRIX * PT)
    return P

# Compute the posterior probability of observing a set of replacements
#
# The integration code demands that the samples are uniformly distributed.
# Numerical integration using simple linear interpolation. 
#
# prePt is a list of pre-computed matrices Pt=expm(Q*t).
def _my_posterior_pre(COUNT_MATRIX, PRE_PT, DIST_SAMPLES):
    L = np.zeros(80)
    
   # Numerical integration, first data point
    P = _log_lik(COUNT_MATRIX, PRE_PT[0])
    P_TOT = P - np.log(2)
    L[0] = P
   
   # middle datapoints
    for i, PRE_PT_ELEMENT in enumerate(PRE_PT[1:]):
        P = _log_lik(COUNT_MATRIX, PRE_PT_ELEMENT)   # log-prob!
        P_TOT = _logprob_add(P_TOT, P)
        L[i+1] = P
   
    # Last datapoint
    P = _log_lik(COUNT_MATRIX, PRE_PT[-1])
    P_TOT = _logprob_add(P_TOT, P - np.log(2))
    L[-1] = P
    
    # 'multiply' each datapoint by sample 'width';
    P_TOT += np.log(DIST_SAMPLES[1] - DIST_SAMPLES[0])
    
    # Setup return value
    POSTERIOR_VEC = np.exp(L - P_TOT)
    return POSTERIOR_VEC


### Interface

# Given an estimate of Q, compute posterior probabilities for all
# distances for all seq pairs. 
#
# Similar to previous comp_posterior, but not re-computing matrix
# exponentials all the time. 
#
def comp_posterior(COUNT_MATRIX_LIST, Q, EQ, DIST_SAMPLES):   
    NUMBER_OF_DIST_SAMPLES = len(DIST_SAMPLES)
    PRE_PT = np.empty(NUMBER_OF_DIST_SAMPLES)
    PRE_PT = np.array([
        np.log(np.diag(EQ) @ scipy.linalg.expm(Q * DIST_SAMPLE))
        for DIST_SAMPLE in DIST_SAMPLES])


    MATRIX_LIST_LENGTH = len(COUNT_MATRIX_LIST)
    PD = np.empty((MATRIX_LIST_LENGTH, NUMBER_OF_DIST_SAMPLES))
    PD = np.array([_my_posterior_pre(COUNT_MATRIX, PRE_PT, DIST_SAMPLES) for COUNT_MATRIX in COUNT_MATRIX_LIST])

    return PD
        


import comp_posterior_JC
import estimate_q
import find_eigens
import matrix_weight
import numpy as np
import simple_estimation

#   Kept for reference

### Interface
def _iterative_estimation(COUNT_MATRIX_LIST, THRESHOLD):
    # These are constant throughout the iterations
    VL, VR, EQ = find_eigens(COUNT_MATRIX_LIST) #   EQ is never changed past this in Octave
    
    # Get a first simple estimate using a Jukes-Cantor model
    distSamples = np.arange(1, 200, 5)
    posterior = comp_posterior_JC(COUNT_MATRIX_LIST, distSamples)
    PW, W = matrix_weight(COUNT_MATRIX_LIST, posterior, distSamplews)
    
    MAX_DIVERGENCE = 100
    Qnew = estimate_q(PW, W, VL, VR, EQ, MAX_DIVERGENCE)
    
    # Use this estimate to as a basis for improvement
    condition = False
    iteration = 0
    dv = []
    MAX_ITERATIONS = 10
    Q = np.matrix((20,20))
    
    while (condition == False):
        iteration += 1
        Q = Qnew
        Qnew, difference = simple_estimation(COUNT_MATRIX_LIST, Q, VL, VR, EQ)
        dv.append(difference)
        
        condition = (iteration >= MAX_ITERATIONS or difference < THRESHOLD)
        
    Q = Qnew
    return Q, EQ, dv